SorhaFrameDrag = LibStub("AceAddon-3.0"):NewAddon("SorhaFrameDrag", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SorhaFrameDrag")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local MODNAME = "SorhaFrameDrag";

BINDING_HEADER_SORHAFRAMEDRAG = "Sorha Frame Drag";
BINDING_NAME_TOGGLE_SORHAFRAMEDRAG = L["Toggle Dragable Frames"]


local tblLazyLoadedFramesToBeLoaded = {}
local tblSetupFrames = {};

local lazyLoadMainFrames = false;
local lazyEnableMainFrames = false;
local lazyDisableMainFrames = false;
local questTrackerInit = false;
local lazyPositionObjectiveTracker = false;


local everDragable = false;

local ignoreSetPoint = false;
local setAttempts = 10;

local tblLazyLoadedFrames = {
	Blizzard_PetJournal = {
		Loaded = false,
		Frame = "PetJournalParent",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_TalentUI = {
		Loaded = false,
		Frame = "PlayerTalentFrame",
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_GuildUI = {
		Loaded = false,
		Frame = "GuildFrame",
		DisableMouseSubframes = {
			"TitleMouseover"
		},	
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	-- Blizzard_StoreUI = { --Too goddamn touchy
	-- 	Loaded = false,
	-- 	Frame = "StoreFrame",
	-- 	InCombat = false,
	-- 	Enabled = false,
	-- },
	Blizzard_EncounterJournal = {
		Loaded = false,
		Frame = "EncounterJournal",
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_AchievementUI = {
		Loaded = false,
		Frame = "AchievementFrame",
		DisableMouseFrames = {
			"AchievementFrameHeader"
		},	
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_BindingUI = {
		Loaded = false,
		Frame = "KeyBindingFrame",
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_TradeSkillUI = {
		Loaded = false,
		Frame = "TradeSkillFrame",
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_MacroUI = {
		Loaded = false,
		Frame = "MacroFrame",
		InCombat = true,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_GuildBankUI = {
		Loaded = false,
		Frame = "GuildBankFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},	
	Blizzard_ItemUpgradeUI = {
		Loaded = false,
		Frame = "ItemUpgradeFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_ItemAlterationUI = {
		Loaded = false,
		Frame = "TransmogrifyFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_TrainerUI = {
		Loaded = false,
		Frame = "ClassTrainerFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_ArchaeologyUI = {
		Loaded = false,
		Frame = "ArchaeologyFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},	
	Blizzard_ItemSocketingUI = {
		Loaded = false,
		Frame = "ItemSocketingFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_BarberShopUI = {
		Loaded = false,
		Frame = "BarberShopFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_Calendar = {
		Loaded = false,
		Frame = "CalendarFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_VoidStorageUI = {
		Loaded = false,
		Frame = "VoidStorageFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_InspectUI = {
		Loaded = false,
		Frame = "InspectFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_AuctionUI = {
		Loaded = false,
		Frame = "AuctionFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
	Blizzard_TimeManager = {
		Loaded = false,
		Frame = "TimeManagerFrame",
		InCombat = false,
		Enabled = false,
		DefaultEnabled = false,
	},
}



local tblDisableMouseFrames = {
	"ReputationFrame",
}


--Defaults
local db
local defaults = {
	profile = {
		DragEnabled = true,
		DebugEnabled = false,
		ObjectiveTrackerDraggable = false,
		ObjectiveTrackerSave = false,
		ObjectiveTrackerLocation = {X = 100, Y = 100, Point = "TOPLEFT", RelativePoint = "TOPLEFT"},
		ObjectiveTrackerHeight = 500,
	},
}

local options = nil
local moduleOptions = {}
local function getOptions()
	if not options then
		options = {
			name = "SorhaFrameDrag ".. "r5",
			type = "group",
			handler = SorhaFrameDrag,
			args = {
				General = {
					name = L["General Options"],
					type = "group",
					order = 1,
					args = {
						DragEnabled = {
							type = "toggle",
							name = L["Enable Frame Drag"],
							width = "full",
							get = function() return db.DragEnabled end,
							set = function(info, value) 
								db.DragEnabled = value;
								if (value == true) then
									SorhaFrameDrag:EnableFrameDrag();
								else
									SorhaFrameDrag:DisableFrameDrag();
								end
							end,
							order = 1,
						},	
						DebugEnabled = {
							type = "toggle",
							width = "full",
							name = L["Enable Debug Messages"],
							get = function() return db.DebugEnabled end,
							set = function(info, value) 
								db.DebugEnabled = value;
							end,
							order = 10,
						},	
						ObjectiveTrackerSpacer = {
							name = "   ",
							width = "full",
							type = "description",
							order = 20,
						},
						ObjectiveTrackerHeader = {
							name = L["Objective Tracker"],
							type = "header",
							order = 21,
						},
						ObjectiveTrackerDraggableEnabled = {
							type = "toggle",
							name = L["Dragable Objective Tracker"],
							width = "full",
							get = function() return db.ObjectiveTrackerDraggable end,
							set = function(info, value) 
								db.ObjectiveTrackerDraggable = value;
								if (value == true) then
									SorhaFrameDrag:EnableDragableObjectiveTracker();
								else
									SorhaFrameDrag:DisableDragableObjectiveTracker();
								end
							end,
							order = 22,
						},
						ObjectiveTrackerSavedEnabled = {
							type = "toggle",
							name = L["Save Objective Tracker Position"],
							width = "full",
							disabled = function() if (Chinchilla) then return true; else return false; end; end,
							get = function() return db.ObjectiveTrackerSave end,
							set = function(info, value) 
								db.ObjectiveTrackerSave = value;
								if (value == true) then
									SorhaFrameDrag:PositionObjectiveTracker();
								end
							end,
							order = 23,
						},
						ObjectiveTrackerHeight = {
							name = L["Objective Tracker Height"],
							desc = L["Set the height of the Objective Tracker"],
							type = 'range',
							disabled = function() if (Chinchilla) then return true; else return false; end; end,
							min = 140,
							max = math.floor(GetScreenHeight()),
							step = 1, bigStep = 5,
							get = function() return db.ObjectiveTrackerHeight end,
							set = function(_, value)
								db.ObjectiveTrackerHeight = value
								ObjectiveTrackerFrame:SetHeight(value)
							end,
							order = 24,
						},
					}
				}
			}
		}
		for k,v in pairs(moduleOptions) do
			options.args[k] = (type(v) == "function") and v() or v
		end
	end

	return options
end

local function hook_UIParentLoadAddOn(name)
	if (SorhaFrameDrag:IsEnabled() == false) then
		return;
	end
	if (tblLazyLoadedFrames[name] == nil) then
		if (db.DebugEnabled == true) then
			print(name .. " NYI");
		end
		return
	end
	if (InCombatLockdown() and tblLazyLoadedFrames[name].InCombat == false) then
		SorhaFrameDrag:RegisterEvent('PLAYER_REGEN_ENABLED');
		tinsert(tblLazyLoadedFramesToBeLoaded, name)
		return
	end
	SorhaFrameDrag:SetupLazyLoadFrame(name);
end

local function hook_ObjectiveTrackerFrameSetPoint()
	if (ignoreSetPoint == true or db.ObjectiveTrackerSave == false) then
		return;
	end
	
	SorhaFrameDrag:PositionObjectiveTracker();	
	ignoreSetPoint = false;
end


local function HandleChatCommand()
	InterfaceOptionsFrame_OpenToCategory(SorhaFrameDrag.optionsFrames.Profiles)
	InterfaceOptionsFrame_OpenToCategory(SorhaFrameDrag.optionsFrames.SorhaFrameDrag)
	InterfaceOptionsFrame:Raise()
end

function SorhaFrameDrag:OnInitialize() 
	self.db = LibStub("AceDB-3.0"):New("SorhaFrameDragDB", defaults, true)
	db = self.db.profile

	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")	
	self:SetupOptions()

	hooksecurefunc("UIParentLoadAddOn", hook_UIParentLoadAddOn)

	self:InitTempDraggableFrames();
end

function SorhaFrameDrag:OnEnable()	
	if (db.DragEnabled == true) then
		self:EnableFrameDrag();
	else
		self:DisableFrameDrag();
	end

	if (db.ObjectiveTrackerDraggable == true) then
		self:EnableDragableObjectiveTracker();
	else
		self:DisableDragableObjectiveTracker();
	end

	if (Chinchilla) then
		db.ObjectiveTrackerSave = false;
	end
	if (db.ObjectiveTrackerSave == true) then
		SorhaFrameDrag:PositionObjectiveTracker();
	end

	self:RegisterEvent("PLAYER_ENTERING_WORLD")
end

function SorhaFrameDrag:PLAYER_ENTERING_WORLD(...)
	self:PositionObjectiveTracker();
end

function SorhaFrameDrag:PLAYER_REGEN_ENABLED(...)
	for key, name in pairs(tblLazyLoadedFramesToBeLoaded) do
		self:SetupLazyLoadFrame(name);
		tblLazyLoadedFramesToBeLoaded[key] = nil;
	end

	if (lazyLoadMainFrames == true) then
		lazyLoadMainFrames = false;
		self:InitTempDraggableFrames();
	end
	if (lazyEnableMainFrames == true) then
		lazyEnableMainFrames = false;
		self:EnableFrameDrag();
	end
	if (lazyDisableMainFrames == true) then
		lazyDisableMainFrames = false;
		self:DisableFrameDrag();
	end

	if (lazyPositionObjectiveTracker == true) then
		lazyPositionObjectiveTracker = false;
		self:PositionObjectiveTracker();
	end

	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
end


function SorhaFrameDrag:SetupOptions() --Creates options. Registers chat commands
	self.optionsFrames = {}
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SorhaFrameDrag", getOptions)
	self.optionsFrames.SorhaFrameDrag = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SorhaFrameDrag", nil, nil, "General")
	
	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), "Profiles")
	
	LibStub("AceConsole-3.0"):RegisterChatCommand("sfd", HandleChatCommand)
	LibStub("AceConsole-3.0"):RegisterChatCommand("sorhaframedrag", HandleChatCommand)
end

function SorhaFrameDrag:RegisterModuleOptions(name, optionTbl, displayName)
	moduleOptions[name] = optionTbl	
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SorhaFrameDrag", displayName or name, "SorhaFrameDrag", name)
end

function SorhaFrameDrag:RefreshConfig()
	db = self.db.profile

	if (db.DragEnabled == true) then
		self:EnableFrameDrag();
	else
		self:DisableFrameDrag();
	end
end



--Frame setups
function SorhaFrameDrag:InitTempDraggableFrames()
	if (InCombatLockdown()) then
		lazyLoadMainFrames = true;
		self:RegisterEvent("PLAYER_REGEN_ENABLED");
		return
	end

	for frameName,info in pairs(UIPanelWindows) do
		if (info.area ~= "full") then
			local frame = _G[frameName]
			if (frame) then
				local titleRegion = frame:GetTitleRegion()
				if (not titleRegion) then
					titleRegion = frame:CreateTitleRegion()
				end
				titleRegion:SetAllPoints(frame)
				
				tinsert(tblSetupFrames, frame);
			end
		end
	end

	--Inventory/bank
	local haveFrames = true;
	local currentContainer = 1;
	while (haveFrames) do
		local containerFrame = _G["ContainerFrame" .. currentContainer];
		
		if (containerFrame) then		
			if (containerFrame) then
				local titleRegion = containerFrame:GetTitleRegion()
				if (not titleRegion) then
					titleRegion = containerFrame:CreateTitleRegion()
				end
				titleRegion:SetAllPoints(containerFrame)
			end			

			local clickTitle = containerFrame.ClickableTitleFrame;
			if (clickTitle) then
				clickTitle:EnableMouse(nil);
			end			

			tinsert(tblSetupFrames, containerFrame);
		else
			haveFrames = false;
		end
		currentContainer = currentContainer + 1;
	end

end

function SorhaFrameDrag:SetupLazyLoadFrame(name)
	if (tblLazyLoadedFrames[name].Loaded == true) then
		return
	end

	local frame = _G[tblLazyLoadedFrames[name].Frame];
	if (frame ~= nil) then
		tblLazyLoadedFrames[name].Loaded = true;

		if (frame:IsMouseEnabled()) then
			tblLazyLoadedFrames[name].DefaultEnabled = true;
		end

		local titleRegion = frame:GetTitleRegion()
		if (not titleRegion) then
			titleRegion = frame:CreateTitleRegion()
		end

		titleRegion:SetAllPoints(frame)

	if (db.DragEnabled == true) then
		self:EnableLazyLoadFrame(name);
	end
		
	end
end


--Frame Enables
function SorhaFrameDrag:EnableFrameDrag()
	if (db.DragEnabled == false) then
		return;
	end
	everDragable = true;

	if (InCombatLockdown()) then
		lazyEnableMainFrames = true;
		self:RegisterEvent("PLAYER_REGEN_ENABLED");
		return
	end

	for i, frame in ipairs(tblSetupFrames) do
		frame:EnableMouse(1)
	end

	for i,frameName in ipairs(tblDisableMouseFrames) do
		local frame = _G[frameName]
		if (frame) then
			frame:EnableMouse(nil)
		end
	end

	for name, tbl in ipairs(tblLazyLoadedFrames) do
		if (tbl.Loaded == true and tbl.Enabled == false) then
			self:EnableLazyLoadFrame(name);
		end
	end

end

function SorhaFrameDrag:DisableFrameDrag()
	if (db.DragEnabled == true or everDragable == false) then
		return;
	end

	if (InCombatLockdown()) then
		lazyDisableMainFrames = true;
		self:RegisterEvent("PLAYER_REGEN_ENABLED");
		return
	end

	for i, frame in ipairs(tblSetupFrames) do
		frame:EnableMouse(nil)
	end

	for i,frameName in ipairs(tblDisableMouseFrames) do
		local frame = _G[frameName]
		if (frame) then
			frame:EnableMouse(1)
		end
	end

	for name, tbl in ipairs(tblLazyLoadedFrames) do
		if (tbl.Loaded == true and tbl.Enabled == true) then
			self:DisableLazyLoadFrame(name);
		end
	end
end


function SorhaFrameDrag:EnableLazyLoadFrame(name)
	local frame = _G[tblLazyLoadedFrames[name].Frame];
	if (frame ~= nil) then
		frame:EnableMouse(1)

		--Kill mouse on subframes
		if (tblLazyLoadedFrames[name].DisableMouseSubframes) then
			for key, value in pairs(tblLazyLoadedFrames[name].DisableMouseSubframes) do
				local frameDisable = frame[value];
				if (frameDisable) then
					frameDisable:EnableMouse(nil)
				end
			end
		end

		--Kill mouse on other frames
		if (tblLazyLoadedFrames[name].DisableMouseFrames) then
			for key, value in pairs(tblLazyLoadedFrames[name].DisableMouseFrames) do
				local frameDisable = _G[value];
				if (frameDisable) then
					frameDisable:EnableMouse(nil)
				end
			end
		end
		everDragable = true;
	end
end

function SorhaFrameDrag:DisableLazyLoadFrame(name)
	local frame = _G[tblLazyLoadedFrames[name].Frame];
	if (frame ~= nil) then
		if (tblLazyLoadedFrames[name].DefaultEnabled == false) then
			frame:EnableMouse(nil)
		end

		--Kill mouse on subframes
		if (tblLazyLoadedFrames[name].DisableMouseSubframes) then
			for key, value in pairs(tblLazyLoadedFrames[name].DisableMouseSubframes) do
				local frameDisable = frame[value];
				if (frameDisable) then
					frameDisable:EnableMouse(1)
				end
			end
		end

		--Kill mouse on other frames
		if (tblLazyLoadedFrames[name].DisableMouseFrames) then
			for key, value in pairs(tblLazyLoadedFrames[name].DisableMouseFrames) do
				local frameDisable = _G[value];
				if (frameDisable) then
					frameDisable:EnableMouse(1)
				end
			end
		end
	end
end



--Quest Tracker
function SorhaFrameDrag:InitDragableObjectiveTracker()
	local frame = _G["ObjectiveTrackerFrame"];
	if (frame ~= nil) then
		self:SecureHook(ObjectiveTrackerFrame, "SetPoint", hook_ObjectiveTrackerFrameSetPoint)
		frame:SetMovable(true)
		frame:SetClampedToScreen(true)
		frame:RegisterForDrag("LeftButton")

		frame:SetScript("OnDragStart", frame.StartMoving)
		frame:SetScript("OnDragStop",  function(self)
			self:StopMovingOrSizing()
			if (db.ObjectiveTrackerSave == true) then
				local strPoint, tempB, strRelativePoint, intPosX, intPosY = self:GetPoint();
				db.ObjectiveTrackerLocation.Point = strPoint
				db.ObjectiveTrackerLocation.RelativePoint = strRelativePoint
				db.ObjectiveTrackerLocation.X = intPosX
				db.ObjectiveTrackerLocation.Y = intPosY				
				SorhaFrameDrag:PositionObjectiveTracker();
			end
		end)

		questTrackerInit = true;
	end
end

function SorhaFrameDrag:PositionObjectiveTracker()
	if (db.ObjectiveTrackerSave == false) then
		return;
	end

	if (InCombatLockdown()) then
		lazyPositionObjectiveTracker = true;
		self:RegisterEvent("PLAYER_REGEN_ENABLED");
		return
	end
	if (questTrackerInit == false) then
		self:InitDragableObjectiveTracker();
	end

	local frame = _G["ObjectiveTrackerFrame"];
	if (frame ~= nil) then
		ignoreSetPoint = true;
		frame:SetHeight(db.ObjectiveTrackerHeight)
		frame:ClearAllPoints()
		frame:SetPoint(db.ObjectiveTrackerLocation.Point, UIParent, db.ObjectiveTrackerLocation.RelativePoint, db.ObjectiveTrackerLocation.X, db.ObjectiveTrackerLocation.Y);
		ignoreSetPoint = false;
	end
end

function SorhaFrameDrag:EnableDragableObjectiveTracker()
	if (questTrackerInit == false) then
		self:InitDragableObjectiveTracker();
	end
	local frame = _G["ObjectiveTrackerFrame"];
	if (frame) then
		frame:EnableMouse(1)
	end
end

function SorhaFrameDrag:DisableDragableObjectiveTracker()
	if (questTrackerInit == true) then
		local frame = _G["ObjectiveTrackerFrame"];
		if (frame) then
			frame:EnableMouse(nil)
		end
	end
end



--Outside Methods
function SorhaFrameDrag:InitCustomFrame(name, frameName)
	local tblNewFrame = {
		Loaded = false,
		Frame = frameName,
		InCombat = false,
		Enabled = false,
	}
	tblLazyLoadedFrames[name]= tblNewFrame;
	self:SetupLazyLoadFrame(name)
end


function SorhaFrameDrag:ToggleFrameDrag()
	db.DragEnabled = not db.DragEnabled ;
	if (db.DragEnabled == true) then
		self:EnableFrameDrag();
	else
		self:DisableFrameDrag();
	end
end

